#!/usr/bin/env python

# Automatically generated by nengo-bones, do not edit this file directly

import io
import pathlib
import runpy

try:
    from setuptools import find_packages, setup
except ImportError:
    raise ImportError(
        "'setuptools' is required but not installed. To install it, "
        "follow the instructions at "
        "https://pip.pypa.io/en/stable/installing/#installing-with-get-pip-py"
    )


def read(*filenames, **kwargs):
    encoding = kwargs.get("encoding", "utf-8")
    sep = kwargs.get("sep", "\n")
    buf = []
    for filename in filenames:
        with io.open(filename, encoding=encoding) as f:
            buf.append(f.read())
    return sep.join(buf)


root = pathlib.Path(__file__).parent
version = runpy.run_path(str(root / "nengo_extras" / "version.py"))["version"]

install_req = [
    "numpy>=1.8",
    "nengo>=3.0",
    "matplotlib>=1.4",
    "pillow",
]
docs_req = [
    "sphinx>=1.8",
    "numpydoc>=0.6",
    "nengo_sphinx_theme>=1.2.0",
    "nbsphinx>=0.8.8",
    "tensorflow",
    "jupyter",
]
optional_req = [
    "nengo-dl",
    "keras",
    "tensorflow",
    "numba>=0.43.1",
    "scipy",
]
tests_req = [
    "jupyter",
    "pytest>=6.0",
    "pytest-allclose",
    "pytest-plt",
    "pytest-rng",
]

setup(
    name="nengo-extras",
    version=version,
    author="Applied Brain Research",
    author_email="info@appliedbrainresearch.com",
    packages=find_packages(),
    url="https://www.nengo.ai/nengo-extras",
    include_package_data=False,
    license="Free for non-commercial use",
    description="Lesser used features for the Nengo neural simulator",
    long_description=read("README.rst", "CHANGES.rst"),
    zip_safe=False,
    install_requires=install_req,
    extras_require={
        "all": docs_req + optional_req + tests_req,
        "docs": docs_req,
        "optional": optional_req,
        "tests": tests_req,
    },
    python_requires=">=3.6",
    classifiers=[
        "Intended Audience :: Science/Research",
        "License :: Free for non-commercial use",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
    ],
)
